/*

    this source file is a part of 
    lp_parse - a preprocessor for MILP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "lp_out.h"
#include "lp_index.h"
#include "lp_tools.h"
#include "lp_parse.h"

int lps_quiet = FALSE;

#define N_FCT_TAGS 4

#define IF_FCT_TAG "If"
#define SUM_FCT_TAG "Sum"
#define FORALL_FCT_TAG "ForAll"
#define INT_TAG "int"

#define IF_FCT 1
#define SUM_FCT 2
#define FORALL_FCT 3
#define INT_OP 4

char *fct_tags[N_FCT_TAGS] = {
  IF_FCT_TAG,
  SUM_FCT_TAG,
  FORALL_FCT_TAG,
  INT_TAG
};

CNR fct_tag_numbers[N_FCT_TAGS] = {
  IF_FCT,
  SUM_FCT,
  FORALL_FCT,
  INT_OP
};

/* read and output to lp_outfile */
int ParseExpression(char **s, CONSTANT *C, INDEX *I, INDEX *LocalIndexList, 
                    int *did_gen_out) {
  char *expr = *s, *tmp1, tmpc1, *tmp2, tmpc2, char_between;
  char *name;
  CONSTANT *c;
  INDEX *i;
  double value;
  int n, dgo;
  CNR fct;
  *did_gen_out = FALSE;
  do {
      
#ifdef DEBUG
    /*printf("pe: %s\n", *s);*/
#endif

    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    if (!**s || (**s == '#')/* || (**s == ';')*/) return TRUE;
    if (**s == '(') {
      tmp1 = scan_end_of_braces(*s);
      if (!tmp1) return FALSE;      
      tmpc1 = *tmp1;
      *tmp1 = '\0';
      ++*s;
      if (!ParseAndEvaluateArithmeticExpression(s, C, 
						LocalIndexList, &value)) 
	return FALSE;
      *tmp1 = tmpc1;
      *s = tmp1 + 1;
      stream_out_number(value);
      *did_gen_out = TRUE;
    } else  if (strchr("+-*/=<>;", **s)) {
      stream_out_char(**s);
      ++*s;
    } else {
      tmp1 = *s;
      SKIP_UNTIL(tmp1, " \t\n\r;#,[](){}+-*/^=<>");
      tmpc1 = *tmp1;
      *tmp1 = '\0';
      n = N_FCT_TAGS;
      fct = getcnr(*s, fct_tags, fct_tag_numbers, &n);
      *tmp1 = tmpc1;
      if (fct) *s += strlen(fct_tags[n]);
      if (fct == INT_OP) {
        stream_out_token(INT_TAG);
        stream_out_char(' ');
        *did_gen_out = TRUE;
      } else
        if (!fct) {

#ifdef DEBUG
	  /*printf("parse symbol %s\n", *s);*/
#endif

	  name = ParseSymbol(s, C, LocalIndexList);
      
#ifdef DEBUG
	  /*printf("parsed symbol %s\n", name);*/
#endif

          if (!name) return FALSE;
          if (strchr(".0123456789", *name)) {
            stream_out_token(name);
          } else {
            c = SearchCONSTANT(C, name);
            if (c  && c->has_value)
              stream_out_number(c->value);
            else {
              i = SearchINDEX(LocalIndexList, name);
              if (i && i->has_value)
                stream_out_number(i->value);
              else {
                c = SearchCONSTANT(Variables, name);
                if (!c && !lps_quiet) 
		  fprintf(mystderr, "%s:%s:%i: %s: undeclared variable\n", 
			  pn, filename(lps_infile->name), lps_infile->cln, 
			  name);
                stream_out_token(name);
		if (strcmp(name, "int") == 0) stream_out_char(' ');
              }
            }
          }
          *did_gen_out = TRUE;
        } else {
          SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
          if (**s != '(') {
            fprintf(mystderr, "%s:%s:%i: %s: expected '('\n", 
		    pn, filename(lps_infile->name), lps_infile->cln, expr);
            return FALSE;
          }
          tmp1 = scan_end_of_braces(*s);
          if (!tmp1) return FALSE;
          tmpc1 = *tmp1;
          *tmp1 = '\0';
          ++*s;
          if (fct == IF_FCT) {
            if (!ParseAndEvaluateArithmeticExpression(s, C, LocalIndexList, 
                                                      &value)) 
              return FALSE;
            SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
            if (**s != ';') {
              fprintf(mystderr, "%s:%s:%i: %s: expected ';'\n", 
		      pn, filename(lps_infile->name), lps_infile->cln, expr);
              return FALSE;
            }
            ++*s;
	    if ((int)(value+0.5)) ParseExpression(s, C, I, LocalIndexList, 
						  did_gen_out);
          } else {
            if (fct == SUM_FCT) char_between = '+'; else char_between = ';';
            SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
            if (**s != '[') {
              fprintf(mystderr, "%s:%s:%i: %s: expected '['\n", 
		      pn, filename(lps_infile->name), lps_infile->cln, expr);
              return FALSE;
            }
            tmp2 = scan_end_of_braces(*s);
            if (!tmp2) return FALSE;
            tmpc2 = *tmp2;
            *tmp2 = '\0';
            ++*s;
            if (!ParseIndexListInFunction(s, C, I, &LocalIndexList, &n)) 
              return FALSE;
#ifdef DEBUG
	    /*printf("back from pilif %i\n", n);*/
#endif
            *tmp2 = tmpc2;
            *s = tmp2 + 1;
            SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
            if (**s != ';') {
              fprintf(mystderr, "%s:%s:%i: %s: expected ';'\n", 
		      pn, filename(lps_infile->name), lps_infile->cln, expr);
              return FALSE;
            }
            ++*s;
	    if (INDEXListToMin2(C, LocalIndexList, n)) {
	      do {

		if (IsIndexListLoop(LocalIndexList, n)) {
#ifdef DEBUG
		  /*printf("function loop %s\n", *s);*/
		  /*PrintIndexes(LocalIndexList);*/
#endif
		  tmp1 = *s;
		  ParseExpression(&tmp1, C, I, LocalIndexList, &dgo);
		  if (dgo) {
		    stream_out_char(char_between);
		    *did_gen_out = TRUE;
		  }
		}
	      } while (LoopINDEXList2(C, LocalIndexList, n));
	    }
	    RemoveFromINDEXList(&LocalIndexList, n);
          }
          *tmp1 = tmpc1;
          *s = tmp1 + 1;
        }
    }
  } while (TRUE);
}

XFILE *datafile = NULL;

int RedirectDatafile(char *filename) {

#ifdef DEBUG
  /*printf("redir: %s\n", filename);*/
#endif

  if (datafile && (datafile->f) && (datafile->f != stdin)) 
    fclose(datafile->f);
  if (*filename == '\0') {
    datafile = xstdin();
  } else {
    if (*filename == '?')
      datafile = NULL;
    else {
      datafile = open_file(filename, "r");
      if (!datafile) return FALSE;      
    }
  }
  return TRUE;
}

int GetData(CONSTANT *C) {
  char line[MAX_LINE_LEN+1], *s;
  while (C) {
    if (!(C->has_value)) {
      if (datafile != NULL) {
        do {
          s = xgetline(datafile, line, MAX_LINE_LEN);
          if (s) SKIP_UNTIL_NOT(s, LP_WHITESPACE);
        } while ((s) && (strchr("#;", *s) != NULL));

#ifdef DEBUG
	/*printf("getdata: %s\n", s);*/
#endif

        if ((!s) || (sscanf(s, "%lf", &(C->value)) < 1)) {
          fprintf(mystderr, "%s:%s:%i: %s: data input error\n", 
		  pn, datafile->name, datafile->cln, s);
          return FALSE;
        }
      } else {
        printf("%s ? ", C->name);
        gets(line);
        if (sscanf(line, "%lf", &(C->value)) < 1) {
          printf("input error, second error quits\n");
          printf("%s ? ", C->name);
          gets(line);
          if (sscanf(line, "%lf", &(C->value)) < 1) {
            return FALSE;
          }
        }
      }
    }
    C->has_value = TRUE;
    C = C->next;
  }
  return TRUE;
}

#define N_TAGS 9

#define CONST_TAG "const:"
#define INDEX_TAG "index:"
#define DATA_TAG "data:"
#define VAR_TAG "var:"
#define MAX_TAG "max:"
#define MIN_TAG "min:"
#define ST_TAG "s.t."
#define INPUT_TAG "input:"
#define DEBUG_TAG "debug:"

#define CONST_SECT 1
#define INDEX_SECT 2
#define DATA_SECT 1
#define VAR_SECT 8
#define MAX_SECT 4
#define MIN_SECT 5
#define ST_SECT 6
#define INPUT_SECT 7
#define DEBUG_SECT 9

char *tags[N_TAGS] = {
  CONST_TAG,
  INDEX_TAG,
  DATA_TAG,
  VAR_TAG,
  MAX_TAG,
  MIN_TAG,
  ST_TAG,
  INPUT_TAG,
  DEBUG_TAG
};

CNR tag_numbers[N_TAGS] = {
  CONST_SECT,
  INDEX_SECT,
  DATA_SECT,
  VAR_SECT,
  MAX_SECT,
  MIN_SECT,
  ST_SECT,
  INPUT_SECT,
  DEBUG_SECT
};

#define MAX_EXTLINE_LEN 10240

char line[MAX_EXTLINE_LEN];

CONSTANT *Constants = NULL, *Variables = NULL;
INDEX *DeclaredIndexes = NULL;

int get_lps_line(char **s, long *section, int *section_change) {
  char *tmp1, tmpc1;
  CNR tmp;
  int i;
  do {
    *s = getextline(lps_infile, line, MAX_EXTLINE_LEN);

#ifdef DEBUG
    /*printf("get line: %s\n", *s);*/
#endif

    if (!(*s)) {
      streamer_flush();
      return FALSE;
    }
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  } while (strchr(";#", **s) != NULL);
  i = N_TAGS;
  tmp1 = *s;
  SKIP_UNTIL(tmp1, " \t\n\r;#,[](){}+-*/^=<>");
  tmpc1 = *tmp1;
  *tmp1 = '\0';
  tmp = getcnr(*s, tags, tag_numbers, &i);

#ifdef DEBUG
  /*printf("get tag %s %li\n", *s, tmp);*/
#endif

  *tmp1 = tmpc1;
  *section_change = FALSE;
  if (tmp) {
    *section = tmp;
    *s += strlen(tags[i]);
    *section_change = TRUE;
  }
  return TRUE;
}

int ParseLPDeclarations(char **s, long *section, GetDataFct gdf) {
  INDEX *i;
  int sc;
  while (get_lps_line(s, section, &sc)) {

    if (*section == DEBUG_SECT) {
      printf("Indexes:\n");
      PrintIndexes(DeclaredIndexes);
      printf("Constants:\n");
      PrintConstant(Constants);
      printf("Variables:\n");
      PrintConstant(Variables);
      exit(1);
    }

    if (sc && (*section != INPUT_SECT)) {
      if (!gdf(Constants)) return FALSE;
    }
    
#ifdef DEBUG
    /*printf("%li: parsing %s\n", *section, *s);*/
#endif

    if (**s) switch (*section) {
    case INDEX_SECT:      
      i = ParseIndexDeclaration(s, Constants, DeclaredIndexes);
      if (i == NULL) return FALSE;
      AddINDEX(&DeclaredIndexes, i);
      break;
    case CONST_SECT:
      if (!ParseConstantDeclaration(s, Constants,DeclaredIndexes, &Constants)) 
        return FALSE;
      break;
    case VAR_SECT:
      if (!ParseConstantDeclaration(s, Constants,DeclaredIndexes, &Variables)) 
        return FALSE;
      break;
    case INPUT_SECT:
      if (!RedirectDatafile(strtok(*s," \t\r\n"))) 
        return FALSE;
      break;
    case MAX_SECT:;
    case MIN_SECT:;
      return TRUE;
    }    
  }
  return TRUE;
}

int ParseLPModel(char **s, long *section) {  
  int did_gen_out, sc;

#ifndef NOHASH
  HashCONSTANT_list(&Constants);
  /*print_hash(Constants);*/
  HashCONSTANT_list(&Variables);
#endif

  do {

#ifdef DEBUG
    /*printf("parsing model line: %s\n", *s);*/
#endif

    did_gen_out = FALSE;
    if (**s) switch (*section) {    
    case MAX_SECT:
      fprintf(lp_outfile->f, "%s ", MAX_TAG);
      if (!ParseExpression(s, Constants, DeclaredIndexes, NULL, &did_gen_out)) 
        return FALSE;
      break;
    case MIN_SECT:
      fprintf(lp_outfile->f, "%s ", MIN_TAG);
      if (!ParseExpression(s, Constants, DeclaredIndexes, NULL, &did_gen_out)) 
        return FALSE;
      break;
    case ST_SECT:
      if (!ParseExpression(s, Constants, DeclaredIndexes, NULL, &did_gen_out)) 
        return FALSE;
      break;
    }
    if (did_gen_out) stream_out_char(';');
  } while (get_lps_line(s, section, &sc));
  return TRUE;
}

int parse_lps() {
  long section;
  char *s;
  RedirectDatafile("");

#ifdef DEBUG
  printf("ParseLPDeclarations:\n");
#endif

  if (!ParseLPDeclarations(&s, &section, GetData)) return FALSE;

#ifdef DEBUG
  printf("ParseLPModel:\n");
#endif

  if (!ParseLPModel(&s, &section)) return FALSE;
  /*fprintf(mystderr, "%s: parsing ok\n", pn);*/
  return TRUE;
}







