/*

    this source file is a part of 
    lp_parse - a preprocessor for MILP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "lp_index.h"
#include "lp_out.h"
#include "lp_tools.h"

#define SKIP_COMMA(s) { \
    if (*(s) != ',') { \
      fprintf(mystderr, "%s:%s:%i: expected ',' after %s\n", \
        pn, filename(lps_infile->name), lps_infile->cln, expr); \
      return FALSE; \
    } \
    (s)++; }

#define BRACKET_ERROR "%s:%s:%i: near %s: missing closing bracket\n", pn, filename(lps_infile->name), lps_infile->cln, expr

/* scans the closing bracket, returns NULL if finds '\0' before it */
char *scan_end_of_braces(char *s) {
    char *expr = s;
    s++;
    do {
        SKIP_UNTIL(s,"{[()]}");
        switch (*s) {
        case '\0':
            fprintf(mystderr, BRACKET_ERROR);
            return NULL;
        case '{':;
        case '[':;
        case '(':
            s = scan_end_of_braces(s);
            if (s) {
                s++;
                break;
            }
            fprintf(mystderr, BRACKET_ERROR);
            return NULL;
        case ')':;
        case ']':;
        case '}':
            return s;
        }
    } while (TRUE);
}

double calculate(double x1, char op, double x2) {
    switch (op) {
    case '+': return x1 + x2;
    case '-': return x1 - x2;
    case '*': return x1 * x2;
    case '/': {if ((x2 > -0.000000000000001) && (x2 < 0.000000000000001))
      fprintf(mystderr, "%s:%s:%i: an attempt to divide by zero: %f/%f", 
	      pn, filename(lps_infile->name), lps_infile->cln, x1, x2);
    else
      return x1 / x2;
    }
    }
    return 0;
}

void reduce(double *x, char *op) {
    switch (op[1]) {
    case '+':;
    case '-':
        x[0] = calculate(x[0], op[0], x[1]);
        op[0] = op[1];
        x[1] = x[2];
        break;
    case '*':;
    case '/':
        x[1] = calculate(x[1], op[1], x[2]);
        break;
    }
}

char *gettok(char **s, char *separators) {
    char *x, tmpc, *tmp, *ret;
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    x = *s;
    SKIP_UNTIL(x, separators);
    tmp = x - 1;
    while (strchr(LP_WHITESPACE, *tmp) != NULL) tmp--;
    tmp++;
    tmpc = *tmp;
    *tmp = '\0';
    ret = xstrdup(*s);
    *tmp = tmpc;
    *s = x;
    return ret;
}

int SkipSymbol(char **s) {
  char *tmp;
  SKIP_UNTIL(*s, " \t\n\r;#,[](){}+-*/^=<>@");
  if (**s == '[') {
    tmp = scan_end_of_braces(*s);
    if (!tmp) return FALSE;      
    *s = tmp + 1;
  }
  return TRUE;
}

int SkipUntilEndOfAE(char **s) {
  char *tmp;
  while (TRUE) {
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    switch (**s) {
    case '(': /* it's a recursion because of braces */
      tmp = scan_end_of_braces(*s);
      if (!tmp) return FALSE;      
      *s = tmp + 1;
      break;
    case '.':;
    case '0':;
    case '1':;
    case '2':;
    case '3':;
    case '4':;
    case '5':;
    case '6':;
    case '7':;
    case '8':;
    case '9': /* it's a number */
      tmp = *s;
      SKIP_UNTIL_NOT(tmp,".0123456789");
      *s = tmp;
      break;
    default: /* it's a name */ 
      if (!SkipSymbol(s)) return FALSE;
    }
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    switch (**s) {
    case '+':;
    case '-':;
    case '*':;
    case '/':
      break;
    default: /* expression ends */
      return TRUE;
    }
    ++*s;
  }
}

int ParseAndEvaluateArithmeticExpression(char **s, CONSTANT *C, INDEX *I, 
					 double *value) {
  int i = 0;
  double x[3];
  char op[2] = {'\0', '\0'}, *tmp, tmpc;
  *value = 0;
  while (TRUE) {
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    switch (**s) {
    case '(': /* it's a recursion because of braces */
      tmp = scan_end_of_braces(*s);
      if (!tmp) return FALSE;      
      tmpc = *tmp;
      *tmp = '\0';
      ++*s;
      if (!ParseAndEvaluateArithmeticExpression(s, C, I, &(x[i]))) 
	return FALSE;
      *tmp = tmpc;
      *s = tmp + 1;
      break;
    case '.':;
    case '0':;
    case '1':;
    case '2':;
    case '3':;
    case '4':;
    case '5':;
    case '6':;
    case '7':;
    case '8':;
    case '9': /* it's a number */
      tmp = *s;
      SKIP_UNTIL_NOT(tmp,".0123456789");
      tmpc = *tmp;
      *tmp = '\0';
      SSCANF_DOUBLE(*s, x[i], FALSE, lps_infile);
      *tmp = tmpc;
      *s = tmp;
      break;
    default: /* it's a name */ 
      if (!ParseAndEvaluateSymbol(s, C, I, &(x[i]))) return FALSE;
    }
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    switch (**s) {
    case '+':;
    case '-':;
    case '*':;
    case '/':
      if (i == 2) {
	reduce(x, op);
	i = 1;
      }
      break;
    default: /* expression ends */
      switch (i) {
      case 0:
	*value = x[0];
	break;
      case 1:
	*value = calculate(x[0], op[0], x[1]);
	break;
      case 2:
	reduce(x, op);
	*value = calculate(x[0], op[0], x[1]);
      }
      return TRUE;
    }
    op[i] = **s;
    ++*s;
    i++;                
  }
}

char *AddIndex(char *name, int index) {
  char itoas[20];
      
#ifdef DEBUG
  /*printf("add index %i %s\n", index, name);*/
#endif

  sprintf(itoas, "_%i", index);
  name = xrealloc(name, strlen(name)+strlen(itoas)+1);
  if (name == NULL) return NULL;
  return strcat(name, itoas);
}

int StripIndexes(char **s, char *buf, int buf_size, int *indexes, int *n) {
  char *expr = *s, *x = *s, tmpc;
  int i = 0, d;
  SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  SKIP_UNTIL(x,"_ \t\r\n");
  d = (*x == '_');
  tmpc = *x;
  *x = '\0';
  buf[0] = '\0';
  strncat(buf, *s, buf_size - 1);
  *x = tmpc;
  while (d && (i < *n)) {
    x++;
    *s = x;
    SKIP_UNTIL(x, "_ \t\r\n");
    d = (*x == '_');
    tmpc = *x;
    *x = '\0';
    SSCANF_INT(*s, indexes[i], FALSE, lps_infile);
    *x = tmpc;
    i++;
  }
  *s = x;
  SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  if ((i == *n) && d) {
    fprintf(mystderr, "%s:%s:%i: near %s: index overflow", 
	    pn, filename(lps_infile->name), lps_infile->cln, expr);
    return FALSE;
  }
  *n = i;
  return TRUE;
}

char *ParseSymbol(char **s, CONSTANT *C, INDEX *I) {
  char *expr = *s, *ret;
  double i_value;
  ret = gettok(s, " \t\n\r;#,[](){}+-*/^=<>@");
  if (**s == '[') {
    ++*s;
    do {
      SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
      if (!**s) {
	fprintf(mystderr, 
		"%s:%s:%i: near %s: expected an index or ']' \n", 
		pn, filename(lps_infile->name), lps_infile->cln, expr);
	free(ret);
	return NULL;
      }
      if (**s == ']') {
	++*s;
	return ret;
      }
      if (!ParseAndEvaluateArithmeticExpression(s, C, I, &i_value)) {
	free(ret);
	return NULL;
      }
      ret = AddIndex(ret, (int)(i_value+0.5));
      SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
      if (**s == ',') ++*s;
    } while (TRUE);
  }
  return ret;
}

int ParseAndEvaluateSymbol(char **s, CONSTANT *C, INDEX *I, double *value) {
  char *name;
  CONSTANT *c;
  INDEX *i;
  name = ParseSymbol(s, C, I);
  if (name == NULL) return FALSE;
  c = SearchCONSTANT(C, name);
  if (c  && c->has_value) {
    *value = c->value;
    return TRUE;
  }
  i = SearchINDEX(I, name);
  if (i && i->has_value) {
    *value = i->value;
    return TRUE;
  }
  if (**s == '\0')
    fprintf(mystderr, "%s:%s:%i: unexpected end of expression near %s\n",
	    pn, filename(lps_infile->name), lps_infile->cln, name);
  else
    fprintf(mystderr, "%s:%s:%i: cannot evaluate symbol ""%s""\n", 
	    pn, filename(lps_infile->name), lps_infile->cln, name);
  return FALSE;
}

int ParseMinAndMaxAEs(char **s, char **min, char **max) {
  char *tmp, tmpc;

#ifdef DEBUG
  /*printf("ParseMinAndMaxAE %s\n", *s);*/
#endif

  tmp = *s;
  SkipUntilEndOfAE(&tmp);
  tmpc = *tmp;
  *tmp = '\0';
  *min = strdup(*s);
  *tmp = tmpc;
  *s = tmp;
  if (**s == ':') {
    ++*s;
    tmp = *s;    
    SkipUntilEndOfAE(&tmp);
    tmpc = *tmp;
    *tmp = '\0';
    *max = strdup(*s);
    *tmp = tmpc;
    *s = tmp;
  } else {
    *max = *min;
  }

#ifdef DEBUG
  /*printf("ParseMinAndMaxAE %s %s\n", *min, *max);*/
#endif

  return TRUE;
}


int ParseMinAndMax(char **s, CONSTANT *C, INDEX *I, int *min, int *max) {
  char *expr = *s, *tmp, tmpc;
  double value;

#ifdef DEBUG
  /*printf("ParseMinAndMax %s\n", *s);*/
#endif

  tmp = *s;
  SKIP_UNTIL(tmp,":");
  if (!*tmp) {
    fprintf(mystderr, "%s:%s:%i: expected ':' near %s\n", 
	    pn, filename(lps_infile->name), lps_infile->cln, expr);
    return FALSE;
  }
  tmpc = *tmp;
  *tmp = '\0';
  if (!ParseAndEvaluateArithmeticExpression(s, C, I, &value)) return FALSE;
  *min = (int)(value + 0.5);
  *tmp = tmpc;
  *s = tmp + 1;
  if (!ParseAndEvaluateArithmeticExpression(s, C, I, &value)) return FALSE;
  *max = (int)(value + 0.5);    
  return TRUE;
}

INDEX *ParseIndexDeclaration(char **s, CONSTANT *C, INDEX *I) {
  INDEX *i;

#ifdef DEBUG
  /*printf("parsing index decl. %s\n", *s);*/
#endif

  i = NewINDEX();
  if (i == NULL) return NULL;
  i->name = gettok(s, " \t:=");
  if (i->name == NULL) {
    free(i);
    return NULL;
  }
  SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  if ((**s == '\0') || (**s == '#')) return i;
  if (!((**s == ':') || (**s == '='))) {
    fprintf(mystderr, "%s:%s:%i: expected ':' or '=' near %s\n", 
	    pn, filename(lps_infile->name), lps_infile->cln, i->name);
    free(i);
    return NULL;
  }
  ++*s;
  if (!ParseMinAndMax(s, C, I, &(i->min), &(i->max))) {
    free(i);
    return NULL;
  }
  i->has_bounds = TRUE;
  return i;
}

int ParseIndexListInConstantDeclaration(char **s, CONSTANT *C, INDEX *I, 
					INDEX **NewI, int *n) {
  INDEX *i, *tmpi;
  char *expr = *s, *tmp, *s2;
  double x;
  *n = 0;
  do {    
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    if (!**s) {
      fprintf(mystderr, "%s:%s:%i: expected index near %s\n", 
	      pn, filename(lps_infile->name), lps_infile->cln, expr);
      return FALSE;
    }
    i = NewINDEX();
    if (!i) return FALSE;
    tmp = *s;
    i->name = gettok(s, ",:=");

#ifdef DEBUG
    /*printf("index in list %s\n", i->name);*/
#endif

    if ((tmpi = SearchINDEX(I, i->name)) != NULL ) {

#ifdef DEBUG
      /*printf("found\n");*/
#endif

      if (tmpi->has_bounds) {
	i->min = tmpi->min;
	i->max = tmpi->max;
	i->has_bounds = TRUE;
      }
      if ((**s == ':') || (**s == '=')) {
	++*s;
	/*
	if (!ParseMinAndMax(s, C, NULL, &(i->min), &(i->max))) {
	  free(i);
	  return FALSE;
	}
	i->has_bounds = TRUE;
	*/
	if (!ParseMinAndMaxAEs(s, &(i->smin), &(i->smax))) {
	  free(i);
	  return FALSE;
	}
	i->has_bounds = FALSE;
      }      
    } else {
      s2 = *s;
      *s = tmp;
      if (*s2 == ':') {
	/*
	if (!ParseMinAndMax(s, C, NULL, &(i->min), &(i->max))) {
	  free(i);
	  return FALSE;
	}
	i->has_bounds = TRUE;
	*/
	if (!ParseMinAndMaxAEs(s, &(i->smin), &(i->smax))) {
	  free(i);
	  return FALSE;
	}
	i->has_bounds = FALSE;
      } else {
	if (!ParseAndEvaluateArithmeticExpression(s, C, NULL, &x)) {
	  free(i);
	  return FALSE;
	}
	i->min = (int)(x+0.5);
	i->max = (int)(x+0.5);
	i->has_bounds = TRUE;
      }
    }
    AddINDEX(NewI, i);
    (*n)++;
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    if (!(**s)) return TRUE;
    SKIP_COMMA(*s);
  } while (TRUE);
}

int INDEXListToMin2(CONSTANT *C, INDEX *I, int n) {
  INDEX *root = I;
  int i, j;
  double value;
  char *s;
      
#ifdef DEBUG
  /*printf("INDEXListToMin2 %i\n", n);*/
#endif

  for (i = n - 1; i >= 0; i--) {
    I = root;
    j = 0;
    while (I && (j < i)) {      
      I = I->next;
      j++;
    }
    if (!I->has_bounds) {
      s = I->smin;
      if (!ParseAndEvaluateArithmeticExpression(&s, C, I->next, &value)) {
	return FALSE;
      }
      I->min = value;
      s = I->smax;
#ifdef DEBUG
      /*printf("p ae %s\n", s);*/
#endif
      if (!ParseAndEvaluateArithmeticExpression(&s, C, I->next, &value)) {
	return FALSE;
      }
#ifdef DEBUG
      /*printf("p ae %f\n", value);*/
#endif
      I->max = value;
    }
    /*if (I->min > I->max) return FALSE;  no loops -> LoopIndexList2 */
    I->has_value = TRUE;
    I->value = I->min;
  }
  return TRUE;
}

int IsIndexListLoop(INDEX *I, int n) {
  int i = 0;
  while (I && (i < n)) {
    if (I->min > I->max) return FALSE;
    I = I->next;
    i++;
  }
  return TRUE;
}

int LoopINDEXList2(CONSTANT *C, INDEX *I, int n) {
  INDEX *root = I;
  int i = 0;
  while (I && (i < n)) {

#ifdef DEBUG
    /*printf("loop indexlist %s %i\n", I->name, I->value);*/
#endif

    if (I->min <= I->max) I->value++;
    if ((I->min > I->max) || (I->value > I->max)) {

      /*I->value = I->min; wait until all done */

      I = I->next;
      i++;
    } else {      
      return INDEXListToMin2(C, root, i);
    }
  }
  return FALSE;
}

int ParseConstantDeclaration(char **s, CONSTANT *C, INDEX *I, 
			     CONSTANT **NewC) {
  CONSTANT *c;
  INDEX *LocalI = NULL, *ix;
  int i, j, n, table_initialization;
  char *expr = *s, *tmp, tmpc, *tmp2;
  char *name;

#ifdef DEBUG
  /*printf("parsing constant decl. %s\n", *s);*/
#endif

  if (!(**s)) return TRUE;
  name = gettok(s, "=[;#");
  if (!name) return FALSE;
  if (**s != '[') {
    c = NewCONSTANT();
    if (!c) return FALSE;
    c->name = name;
    if (**s != '=') {
      AddCONSTANT(NewC, c);
      return TRUE;
    }
    ++*s;
    if (!ParseAndEvaluateArithmeticExpression(s, C, NULL, &(c->value))) {
      free(c);
      return FALSE;
    }
    c->has_value = TRUE;
    AddCONSTANT(NewC, c);
    return TRUE;
  }
  tmp = scan_end_of_braces(*s);
  if (!tmp) return FALSE;
  tmpc = *tmp;
  *tmp = '\0';
  ++*s;
  if (!ParseIndexListInConstantDeclaration(s, C, I, &LocalI, &n)) {
    DeleteINDEXList(&LocalI);
    return FALSE;
  }
  *tmp = tmpc;
  *s = tmp + 1;
  SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  table_initialization = (**s == '=');
  if (table_initialization) {
    ++*s;
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    if (**s != '{') {
      fprintf(mystderr, "%s:%s:%i: expected '{' near %s\n", 
	      pn, filename(lps_infile->name), lps_infile->cln, expr);
      return FALSE;
    }
    tmp = scan_end_of_braces(*s);
    if (!tmp) return FALSE;
    tmpc = *tmp;
    *tmp = '\0';
    ++*s;
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
  }
  if (INDEXListToMin2(C, LocalI, n)) {
    do {

      if (IsIndexListLoop(LocalI, n)) {
      
#ifdef DEBUG
	/*printf("loop %s %i\n", name, n);*/
	/*PrintIndexes(LocalI);*/
#endif

	c = NewCONSTANT();
	if (!c) return FALSE;
	c->name = xstrdup(name);
	if (!c->name) return FALSE;
	for (i = 0; i < n; i++) {
	  ix = LocalI;
	  for (j = i + 1; j < n; j++) ix = ix->next;
	  c->name = AddIndex(c->name, ix->value);
	}
	if (table_initialization) {
	  tmp2 = *s;
	  if (!ParseAndEvaluateArithmeticExpression(s, C, LocalI, 
						    &(c->value))) {
	    free(c);
	    DeleteINDEXList(&LocalI);
	    return FALSE;
	  }
	  c->has_value = TRUE;
	  SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
	  if (!(**s)) *s = tmp2; else SKIP_COMMA(*s);
	}
	AddCONSTANT(NewC, c);
      
#ifdef DEBUG
	/*printf("new constant %s\n", c->name);*/
#endif

      }

    } while (LoopINDEXList2(C, LocalI, n));
  }
  if (table_initialization) {
    *tmp = tmpc;
    *s = tmp + 1;
  }
  return TRUE;
}

/*---------------------------------------------------------------------------*/

int ParseIndexListInFunction(char **s, CONSTANT *C, INDEX *I, 
                             INDEX **NewI, int *n) {
  INDEX *i, *tmpi;
  char *expr = *s, *tmp;
  *n = 0;
  do {
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    i = NewINDEX();
    if (!i) return FALSE;    
    tmp = *s;
    i->name = gettok(s, ",:=");
    if (!i->name) return FALSE;    
    if (strchr(",", **s)) {
      if ((tmpi = SearchINDEX(I, i->name)) == NULL ) {
	free(i);
	fprintf(mystderr, 
		"%s:%s:%i: undeclared index in function near %s\n",
		pn, filename(lps_infile->name), lps_infile->cln, expr);
	return FALSE;
      }
      i->min = tmpi->min;
      i->max = tmpi->max;
      i->has_bounds = TRUE;
    } else {
      ++*s;
      /*
      if (!ParseMinAndMax(s, C, NULL, &(i->min), &(i->max))) {
	free(i);
	return FALSE;
      }
      */
      if (!ParseMinAndMaxAEs(s, &(i->smin), &(i->smax))) {
	free(i);
	return FALSE;
      }
      i->has_bounds = FALSE;
    }
    AddINDEX(NewI, i);

#ifdef DEBUG
    /*printf("pilif %s %s %i\n", *s, i->name, *n);*/
#endif

    ++*n;
    SKIP_UNTIL_NOT(*s, LP_WHITESPACE);
    if (!(**s)) return TRUE;
    SKIP_COMMA(*s);
  } while (TRUE);
}
